import itertools
from collections import OrderedDict
import numpy as np
from repro_eval.config import TRIM_THRESH, exclude


def trim(run, thresh=TRIM_THRESH):
    """
    Use this function to trim a run to a length of a document length specified by thresh.

    @param run: The run to be trimmed.
    @param thresh: The threshold value of the run length.
    """
    for topic, docs in run.items():
        run[topic] = dict(list(run[topic].items())[:thresh])


def arp(topic_scores):
    """
    This function computes the Average Retrieval Performance (ARP) according to the following paper:
    Timo Breuer, Nicola Ferro, Norbert Fuhr, Maria Maistro, Tetsuya Sakai, Philipp Schaer, Ian Soboroff.
    How to Measure the Reproducibility of System-oriented IR Experiments.
    Proceedings of SIGIR, pages 349-358, 2020.

    The ARP score is defined by the mean across the different topic scores of a run.

    @param topic_scores: Topic scores of an evaluated run.
    @return: The ARP score.
    """
    return np.array(list(topic_scores.values())).mean()


def _arp_scores(run):
    """
    Helping function returning a generator for determining the Average Retrieval Performance (ARP) scores.

    @param run: The run to be evaluated.
    @return: Generator with ARP scores for each trec_eval evaluation measure.
    """
    measures_all = list(list(run.values())[0].keys())
    measures_valid = [m for m in measures_all if m not in exclude]
    topics = run.keys()

    for measure in measures_valid:
        yield measure, np.array(list([run.get(topic).get(measure) for topic in topics])).mean()


def arp_scores(run):
    """
    This function computes the Average Retrieval Performance (ARP) scores according to the following paper:
    Timo Breuer, Nicola Ferro, Norbert Fuhr, Maria Maistro, Tetsuya Sakai, Philipp Schaer, Ian Soboroff.
    How to Measure the Reproducibility of System-oriented IR Experiments.
    Proceedings of SIGIR, pages 349-358, 2020.

    The ARP score is defined by the mean across the different topic scores of a run.
    For all measures outputted by trec_eval, the ARP scores will be determined.

    @param run: The run to be evaluated.
    @return: Dictionary containing the ARP scores for every measure outputted by trec_eval.
    """
    return dict(_arp_scores(run))


def _topic_scores(run_scores):
    """
    Helping function returning a generator for determining the topic scores for each measure.

    @param run_scores: The run scores of the previously evaluated run.
    @return: Generator with topic scores for each trec_eval evaluation measure.
    """
    measures_all = list(list(run_scores.values())[0].keys())
    measures_valid = [m for m in measures_all if m not in exclude]
    topics = run_scores.keys()

    for measure in measures_valid:
        yield measure, [run_scores.get(topic).get(measure) for topic in topics]


def topic_scores(run_scores):
    """
    Use this function for a dictionary that contains the topic scores for each measure outputted by trec_eval.

    @param run_scores: The run scores of the previously evaluated run.
    @return: Dictionary containing the topic scores for every measure outputted by trec_eval.
    """
    return dict(_topic_scores(run_scores))


def print_base_adv(measure_topic, repro_measure, base_value, adv_value=None):
    """
    Pretty print output in trec_eval inspired style. Use this for printing baseline and/or advanced results.

    @param measure_topic: The topic number.
    @param repro_measure: Name of the reproduction/replication measure.
    @param base_value: Value of the evaluated baseline run.
    @param adv_value: Value of the evaluated advanced run.
    """
    if adv_value:
        fill = ('{:3s}' if base_value < 0 else '{:4s}')
        print(('{:25s}{:8s}{:8s}{:.4f}' + fill + '{:8s}{:.4f}').format(measure_topic, repro_measure,
                                                                       'BASE', base_value, ' ', 'ADV', adv_value))
    else:
        print('{:25s}{:8s}{:8s}{:.4f}'.format(measure_topic, repro_measure, 'BASE', base_value))


def print_simple_line(measure, repro_measure, value):
    """
    Use this for printing lines with trec_eval and reproduction/replication measures.
    Pretty print output in trec_eval inspired style.
    @param measure: Name of the trec_eval measure.
    @param repro_measure: Name of the reproduction/replication measure.
    @param value: Value of the evaluated run.
    @return:
    """
    print('{:25s}{:8s}{:.4f}'.format(measure, repro_measure, value))


def break_ties(run):
    """
    Use this function to break score ties like it is implemented in trec_eval.
    Documents with the same score will be sorted in reverse alphabetical order.
    :param run: Run with score ties. Nested dictionary structure (cf. pytrec_eval)
    :return: Reordered run
    """
    for topic, ranking in run.items():
        docid_score_tuple = list(ranking.items())
        reordered_ranking = []
        for k, v in itertools.groupby(docid_score_tuple, lambda item: item[1]):
            reordered_ranking.extend(sorted(v, reverse=True))
        run[topic] = OrderedDict(reordered_ranking)
    return run



import pytrec_eval
from repro_eval.util import trim, break_ties
from repro_eval.measure.statistics import ttest
from repro_eval.measure.overall_effects import ER, deltaRI
from repro_eval.measure.document_order import ktau_union as ktu, RBO
from repro_eval.measure.effectiveness import rmse as RMSE, nrmse as nRMSE
from repro_eval.config import ERR_MSG


class Evaluator(object):
    """
    An abstract evaluator that holds the original baseline and advanced run as well as
    the reproduced/replicated baseline and advanced run.
    """

    def __init__(self, **kwargs):
        self.qrel_orig_path = kwargs.get('qrel_orig_path', None)
        self.run_b_orig_path = kwargs.get('run_b_orig_path', None)
        self.run_a_orig_path = kwargs.get('run_a_orig_path', None)
        self.run_b_rep_path = kwargs.get('run_b_rep_path', None)
        self.run_a_rep_path = kwargs.get('run_a_rep_path', None)
        self.run_b_orig = None
        self.run_a_orig = None
        self.run_b_rep = None
        self.run_a_rep = None
        self.run_b_orig_score = None
        self.run_a_orig_score = None
        self.run_b_rep_score = None
        self.run_a_rep_score = None

        if self.qrel_orig_path:
            with open(self.qrel_orig_path, 'r') as f_qrel:
                qrel_orig = pytrec_eval.parse_qrel(f_qrel)
                self.rel_eval = pytrec_eval.RelevanceEvaluator(qrel_orig, pytrec_eval.supported_measures)

        if self.run_b_orig_path:
            with open(self.run_b_orig_path, 'r') as f_run:
                self.run_b_orig = pytrec_eval.parse_run(f_run)
                self.run_b_orig = {t: self.run_b_orig[t] for t in sorted(self.run_b_orig)}

        if self.run_a_orig_path:
            with open(self.run_a_orig_path, 'r') as f_run:
                self.run_a_orig = pytrec_eval.parse_run(f_run)
                self.run_a_orig = {t: self.run_a_orig[t] for t in sorted(self.run_a_orig)}

        if self.run_b_rep_path:
            with open(self.run_b_rep_path, 'r') as f_run:
                self.run_b_rep = pytrec_eval.parse_run(f_run)
                self.run_b_rep = {t: self.run_b_rep[t] for t in sorted(self.run_b_rep)}

        if self.run_a_rep_path:
            with open(self.run_a_rep_path, 'r') as f_run:
                self.run_a_rep = pytrec_eval.parse_run(f_run)
                self.run_a_rep = {t: self.run_a_rep[t] for t in sorted(self.run_a_rep)}

    def trim(self, t=None, run=None):
        """
        Trims all runs of the Evaluator to the length specified by the threshold value t.

        @param t: Threshold parameter or number of top-k documents to be considered.
        @param run: If run is not None, only the provided run will be trimmed.
        """
        if run:
            run = break_ties(run)
            if t:
                trim(run, thresh=t)
            else:
                trim(run)
            return

        if self.run_b_orig:
            self.run_b_orig = break_ties(self.run_b_orig)
            if t:
                trim(self.run_b_orig, thresh=t)
            else:
                trim(self.run_b_orig)

        if self.run_a_orig:
            self.run_a_orig = break_ties(self.run_a_orig)
            if t:
                trim(self.run_a_orig, thresh=t)
            else:
                trim(self.run_a_orig)

        if self.run_b_rep:
            self.run_b_rep = break_ties(self.run_b_rep)
            if t:
                trim(self.run_b_rep, thresh=t)
            else:
                trim(self.run_b_rep)

        if self.run_a_rep:
            self.run_a_rep = break_ties(self.run_a_rep)
            if t:
                trim(self.run_a_rep, thresh=t)
            else:
                trim(self.run_a_rep)

    def evaluate(self, run=None):
        """
        Evaluates the original baseline and advanced run if available.

        @param run: Reproduced or replicated run that will be evaluated.
        """
        if self.run_b_orig:
            self.run_b_orig = break_ties(self.run_b_orig)
            self.run_b_orig_score = self.rel_eval.evaluate(self.run_b_orig)
        if self.run_a_orig:
            self.run_a_orig = break_ties(self.run_a_orig)
            self.run_a_orig_score = self.rel_eval.evaluate(self.run_a_orig)

    def er(self, run_b_score=None, run_a_score=None, run_b_path=None, run_a_path=None, print_feedback=False):
        """
        Determines the Effect Ratio (ER) according to the following paper:
        Timo Breuer, Nicola Ferro, Norbert Fuhr, Maria Maistro, Tetsuya Sakai, Philipp Schaer, Ian Soboroff.
        How to Measure the Reproducibility of System-oriented IR Experiments.
        Proceedings of SIGIR, pages 349-358, 2020.

        The ER value is determined by the ratio between the mean improvements
        of the original and reproduced/replicated experiments.

        @param run_b_score: Scores of the baseline run,
                            if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_a_score: Scores of the advanced run,
                            if not provided the scores of the RpdEvaluator object will be used instead.
        @param print_feedback: Boolean value indicating if feedback on progress should be printed.
        @return: Dictionary containing the ER values for the specified run combination.
        """
        if print_feedback:
            print('Determining Effect Ratio (ER)')

        if self.run_b_orig_score and self.run_a_orig_score and run_b_path and run_a_path:
            with open(run_b_path, 'r') as b_run, open(run_a_path, 'r') as a_run:
                run_b_rep = pytrec_eval.parse_run(b_run)
                run_b_rep = {t: run_b_rep[t] for t in sorted(run_b_rep)}
                run_b_rep_score = self.rel_eval_rpl.evaluate(run_b_rep) if hasattr(self, 'rel_eval_rpl') else self.rel_eval.evaluate(run_b_rep)
                run_a_rep = pytrec_eval.parse_run(a_run)
                run_a_rep = {t: run_a_rep[t] for t in sorted(run_a_rep)}
                run_a_rep_score = self.rel_eval_rpl.evaluate(run_a_rep) if hasattr(self, 'rel_eval_rpl') else self.rel_eval.evaluate(run_a_rep)
            return ER(orig_score_b=self.run_b_orig_score, orig_score_a=self.run_a_orig_score,
                      rep_score_b=run_b_rep_score, rep_score_a=run_a_rep_score, pbar=print_feedback)

        if self.run_b_orig_score and self.run_a_orig_score and run_b_score and run_a_score:
            return ER(orig_score_b=self.run_b_orig_score, orig_score_a=self.run_a_orig_score,
                      rep_score_b=run_b_score, rep_score_a=run_a_score, pbar=print_feedback)

        if self.run_b_orig_score and self.run_a_orig_score and self.run_b_rep_score and self.run_a_rep_score:
            return ER(orig_score_b=self.run_b_orig_score, orig_score_a=self.run_a_orig_score,
                      rep_score_b=self.run_b_rep_score, rep_score_a=self.run_a_rep_score, pbar=print_feedback)
        else:
            print(ERR_MSG)

    def dri(self, run_b_score=None, run_a_score=None, run_b_path=None, run_a_path=None, print_feedback=False):
        """
        Determines the Delta Relative Improvement (DeltaRI) according to the following paper:
        Timo Breuer, Nicola Ferro, Norbert Fuhr, Maria Maistro, Tetsuya Sakai, Philipp Schaer, Ian Soboroff.
        How to Measure the Reproducibility of System-oriented IR Experiments.
        Proceedings of SIGIR, pages 349-358, 2020.

        The DeltaRI value is determined by the difference between the relative improvements
        of the original and reproduced/replicated experiments.

        @param run_b_score: Scores of the baseline run,
                            if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_a_score: Scores of the advanced run,
                            if not provided the scores of the RpdEvaluator object will be used instead.
        @param print_feedback: Boolean value indicating if feedback on progress should be printed.
        @return: Dictionary containing the DRI values for the specified run combination.
        """
        if print_feedback:
            print('Determining Delta Relative Improvement (DRI)')

        if self.run_b_orig_score and self.run_a_orig_score and run_b_path and run_a_path:
            with open(run_b_path, 'r') as b_run, open(run_a_path, 'r') as a_run:
                run_b_rep = pytrec_eval.parse_run(b_run)
                run_b_rep = {t: run_b_rep[t] for t in sorted(run_b_rep)}
                run_b_rep_score = self.rel_eval_rpl.evaluate(run_b_rep) if hasattr(self, 'rel_eval_rpl') else self.rel_eval.evaluate(run_b_rep)
                run_a_rep = pytrec_eval.parse_run(a_run)
                run_a_rep = {t: run_a_rep[t] for t in sorted(run_a_rep)}
                run_a_rep_score = self.rel_eval_rpl.evaluate(run_a_rep) if hasattr(self, 'rel_eval_rpl') else self.rel_eval.evaluate(run_a_rep)
            return deltaRI(orig_score_b=self.run_b_orig_score, orig_score_a=self.run_a_orig_score,
                           rep_score_b=run_b_rep_score, rep_score_a=run_a_rep_score, pbar=print_feedback)

        if self.run_b_orig_score and self.run_a_orig_score and run_b_score and run_a_score:
            return deltaRI(orig_score_b=self.run_b_orig_score, orig_score_a=self.run_a_orig_score,
                           rep_score_b=run_b_score, rep_score_a=run_a_score, pbar=print_feedback)

        if self.run_b_orig_score and self.run_a_orig_score and self.run_b_rep_score and self.run_a_rep_score:
            return deltaRI(orig_score_b=self.run_b_orig_score, orig_score_a=self.run_a_orig_score,
                           rep_score_b=self.run_b_rep_score, rep_score_a=self.run_a_rep_score, pbar=print_feedback)
        else:
            print(ERR_MSG)

    def _ttest(self, rpd=True, run_b_score=None, run_a_score=None, print_feedback=False):
        """
        Conducts either a paired (reproducibility) or unpaired (replicability) two-sided t-test according to the following paper:
        Timo Breuer, Nicola Ferro, Norbert Fuhr, Maria Maistro, Tetsuya Sakai, Philipp Schaer, Ian Soboroff.
        How to Measure the Reproducibility of System-oriented IR Experiments.
        Proceedings of SIGIR, pages 349-358, 2020.

        @param rpd: Boolean indicating if the evaluated runs are reproduced.
        @param run_b_score: Scores of the baseline run,
                            if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_a_score: Scores of the advanced run,
                            if not provided the scores of the RpdEvaluator object will be used instead.
        @param print_feedback: Boolean value indicating if feedback on progress should be printed.
        @return: Dictionary with p-values that compare the score distributions of the baseline and advanced run.
        """
        if self.run_b_orig_score and (self.run_b_rep_score or run_b_score):
            if run_b_score and run_a_score:
                if print_feedback:
                    print('Determining p-values of t-test for baseline and advanced run.')
                return {'baseline': ttest(self.run_b_orig_score, run_b_score, rpd=rpd, pbar=print_feedback),
                        'advanced': ttest(self.run_a_orig_score, run_a_score, rpd=rpd, pbar=print_feedback)}
            if run_b_score:
                if print_feedback:
                    print('Determining p-values of t-test for baseline run.')
                return {'baseline': ttest(self.run_b_orig_score, run_b_score, rpd=rpd, pbar=print_feedback)}
            if self.run_a_orig_score and self.run_a_rep_score:
                if print_feedback:
                    print('Determining p-values of t-test for baseline and advanced run.')
                return {'baseline': ttest(self.run_b_orig_score, self.run_b_rep_score, rpd=rpd, pbar=print_feedback),
                        'advanced': ttest(self.run_a_orig_score, self.run_a_rep_score, rpd=rpd, pbar=print_feedback)}
            else:
                if print_feedback:
                    print('Determining p-values of t-test for baseline run.')
                return {'baseline': ttest(self.run_b_orig_score, self.run_b_rep_score, rpd=rpd, pbar=print_feedback)}
        else:
            print(ERR_MSG)


class RpdEvaluator(Evaluator):
    """
    The Reproducibility Evaluator is used for quantifying the different levels of reproduction for runs that were
    derived from the same test collection used in the original experiment.
    """

    def evaluate(self, run=None):
        """
        Evaluates the scores of the original and reproduced baseline and advanced runs.
        If a (reproduced) run is provided only this one will be evaluated and a dictionary with the corresponding
        scores is returned.
        @param run: A reproduced run. If not specified, the original and reproduced runs of the the RpdEvaluator will
                    be used instead.
        @return: If run is specified, a dictionary with the corresponding scores is returned.
        """
        if run:
            return self.rel_eval.evaluate(run)

        super(RpdEvaluator, self).evaluate()
        if self.run_b_rep:
            self.run_b_rep = break_ties(self.run_b_rep)
            self.run_b_rep_score = self.rel_eval.evaluate(self.run_b_rep)
        if self.run_a_rep:
            self.run_a_rep = break_ties(self.run_a_rep)
            self.run_a_rep_score = self.rel_eval.evaluate(self.run_a_rep)

    def ktau_union(self, run_b_rep=None, run_a_rep=None, run_b_path=None, run_a_path=None, print_feedback=False):
        """
        Determines Kendall's tau Union (KTU) between the original and reproduced document orderings
        according to the following paper:
        Timo Breuer, Nicola Ferro, Norbert Fuhr, Maria Maistro, Tetsuya Sakai, Philipp Schaer, Ian Soboroff.
        How to Measure the Reproducibility of System-oriented IR Experiments.
        Proceedings of SIGIR, pages 349-358, 2020.

        @param run_b_rep: Scores of the baseline run,
                          if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_a_rep: Scores of the advanced run,
                          if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_b_path: Path to another reproduced baseline run,
                           if not provided the reproduced baseline run of the RpdEvaluator object will be used instead.
        @param run_a_path: Path to another reproduced advanced run,
                           if not provided the reproduced advanced run of the RpdEvaluator object will be used instead.
        @param print_feedback: Boolean value indicating if feedback on progress should be printed.
        @return: Dictionary with KTU values that compare the document orderings of the original and reproduced runs.
        """
        if self.run_b_orig and run_b_path:
            if self.run_a_orig and run_a_path:
                if print_feedback:
                    print("Determining Kendall's tau Union (KTU) for baseline and advanced run.")
                with open(run_b_path, 'r') as b_run, open(run_a_path, 'r') as a_run:
                    run_b_rep = pytrec_eval.parse_run(b_run)
                    run_b_rep = {t: run_b_rep[t] for t in sorted(run_b_rep)}
                    run_a_rep = pytrec_eval.parse_run(a_run)
                    run_a_rep = {t: run_a_rep[t] for t in sorted(run_a_rep)}
                return {'baseline': ktu(self.run_b_orig, run_b_rep, pbar=print_feedback),
                        'advanced': ktu(self.run_a_orig, run_a_rep, pbar=print_feedback)}
            else:
                if print_feedback:
                    print("Determining Kendall's tau Union (KTU) for baseline run.")
                with open(run_b_path, 'r') as b_run:
                    run_b_rep = pytrec_eval.parse_run(b_run)
                    run_b_rep = {t: run_b_rep[t] for t in sorted(run_b_rep)}
                return {'baseline': ktu(self.run_b_orig, run_b_rep, pbar=print_feedback)}

        if self.run_b_orig and run_b_rep:
            if self.run_a_orig and run_a_rep:
                if print_feedback:
                    print("Determining Kendall's tau Union (KTU) for baseline and advanced run.")
                return {'baseline': ktu(self.run_b_orig, run_b_rep, pbar=print_feedback),
                        'advanced': ktu(self.run_a_orig, run_a_rep, pbar=print_feedback)}
            else:
                if print_feedback:
                    print("Determining Kendall's tau Union (KTU) for baseline run.")
                return {'baseline':  ktu(self.run_b_orig, run_b_rep, pbar=print_feedback)}

        if self.run_b_orig and self.run_b_rep:
            if self.run_a_orig and self.run_a_rep:
                if print_feedback:
                    print("Determining Kendall's tau Union (KTU) for baseline and advanced run.")
                return {'baseline': ktu(self.run_b_orig, self.run_b_rep, pbar=print_feedback),
                        'advanced': ktu(self.run_a_orig, self.run_a_rep, pbar=print_feedback)}
            else:
                if print_feedback:
                    print("Determining Kendall's tau Union (KTU) for baseline run.")
                return {'baseline':  ktu(self.run_b_orig, self.run_b_rep, pbar=print_feedback)}
        else:
            print(ERR_MSG)

    def rbo(self, run_b_rep=None, run_a_rep=None, run_b_path=None, run_a_path=None, print_feedback=False, misinfo=True):
        """
        Determines the Rank-Biased Overlap (RBO) between the original and reproduced document orderings
        according to the following paper:
        Timo Breuer, Nicola Ferro, Norbert Fuhr, Maria Maistro, Tetsuya Sakai, Philipp Schaer, Ian Soboroff.
        How to Measure the Reproducibility of System-oriented IR Experiments.
        Proceedings of SIGIR, pages 349-358, 2020.

        @param run_b_rep: Scores of the baseline run,
                          if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_a_rep: Scores of the advanced run,
                          if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_b_path: Path to another reproduced baseline run,
                           if not provided the reproduced baseline run of the RpdEvaluator object will be used instead.
        @param run_a_path: Path to another reproduced advanced run,
                           if not provided the reproduced advanced run of the RpdEvaluator object will be used instead.
        @param print_feedback: Boolean value indicating if feedback on progress should be printed.
        @param misinfo: Use the RBO implementation that is also used in the TREC Health Misinformation Track.
                        See also: https://github.com/claclark/Compatibility
        @return: Dictionary with RBO values that compare the document orderings of the original and reproduced runs.
        """
        if self.run_b_orig and run_b_path:
            if self.run_a_orig and run_a_path:
                if print_feedback:
                    print("Determining Rank-biased Overlap (RBO) for baseline and advanced run.")
                with open(run_b_path, 'r') as b_run, open(run_a_path, 'r') as a_run:
                    run_b_rep = pytrec_eval.parse_run(b_run)
                    run_b_rep = {t: run_b_rep[t] for t in sorted(run_b_rep)}
                    run_a_rep = pytrec_eval.parse_run(a_run)
                    run_a_rep = {t: run_a_rep[t] for t in sorted(run_a_rep)}
                return {'baseline': RBO(self.run_b_orig, run_b_rep, pbar=print_feedback, misinfo=misinfo),
                        'advanced': RBO(self.run_a_orig, run_a_rep, pbar=print_feedback, misinfo=misinfo)}
            else:
                if print_feedback:
                    print("Determining Rank-biased Overlap (RBO) for baseline run.")
                with open(run_b_path, 'r') as b_run:
                    run_b_rep = pytrec_eval.parse_run(b_run)
                    run_b_rep = {t: run_b_rep[t] for t in sorted(run_b_rep)}
                return {'baseline': RBO(self.run_b_orig, run_b_rep, pbar=print_feedback, misinfo=misinfo)}

        if self.run_b_orig and run_b_rep:
            if self.run_a_orig and run_a_rep:
                if print_feedback:
                    print("Determining Rank-biased Overlap (RBO) for baseline and advanced run.")
                return {'baseline': RBO(self.run_b_orig, run_b_rep, pbar=print_feedback, misinfo=misinfo),
                        'advanced': RBO(self.run_a_orig, run_a_rep, pbar=print_feedback, misinfo=misinfo)}
            else:
                if print_feedback:
                    print("Determining Rank-biased Overlap (RBO) for baseline run.")
                return {'baseline':  RBO(self.run_b_orig, run_b_rep, pbar=print_feedback, misinfo=misinfo)}
        if self.run_b_orig and self.run_b_rep:
            if self.run_a_orig and self.run_a_rep:
                if print_feedback:
                    print("Determining Rank-biased Overlap (RBO) for baseline and advanced run.")
                return {'baseline': RBO(self.run_b_orig, self.run_b_rep, pbar=print_feedback, misinfo=misinfo),
                        'advanced': RBO(self.run_a_orig, self.run_a_rep, pbar=print_feedback, misinfo=misinfo)}
            else:
                if print_feedback:
                    print("Determining Rank-biased Overlap (RBO) for baseline run.")
                return {'baseline':  RBO(self.run_b_orig, self.run_b_rep, pbar=print_feedback, misinfo=misinfo)}
        else:
            print(ERR_MSG)

    def rmse(self, run_b_score=None, run_a_score=None, run_b_path=None, run_a_path=None, print_feedback=False):
        """
        Determines the Root Mean Square Error (RMSE) according to the following paper:
        Timo Breuer, Nicola Ferro, Norbert Fuhr, Maria Maistro, Tetsuya Sakai, Philipp Schaer, Ian Soboroff.
        How to Measure the Reproducibility of System-oriented IR Experiments.
        Proceedings of SIGIR, pages 349-358, 2020.

        @param run_b_score: Scores of the baseline run,
                            if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_a_score: Scores of the advanced run,
                            if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_b_path: Path to another reproduced baseline run,
                           if not provided the reproduced baseline run of the RpdEvaluator object will be used instead.
        @param run_a_path: Path to another reproduced advanced run,
                           if not provided the reproduced advanced run of the RpdEvaluator object will be used instead.
        @param print_feedback: Boolean value indicating if feedback on progress should be printed.
        @return: Dictionary with RMSE values that measure the closeness
                 between the topics scores of the original and reproduced runs.
        """
        if self.run_b_orig and run_b_path:
            if self.run_a_orig and run_a_path:
                if print_feedback:
                    print("Determining Root Mean Square Error (RMSE) for baseline and advanced run.")
                with open(run_b_path, 'r') as b_run, open(run_a_path, 'r') as a_run:
                    run_b_rep = pytrec_eval.parse_run(b_run)
                    run_b_rep = {t: run_b_rep[t] for t in sorted(run_b_rep)}
                    run_b_rep_score = self.rel_eval.evaluate(run_b_rep)
                    run_a_rep = pytrec_eval.parse_run(a_run)
                    run_a_rep = {t: run_a_rep[t] for t in sorted(run_a_rep)}
                    run_a_rep_score = self.rel_eval.evaluate(run_a_rep)
                return {'baseline': RMSE(self.run_b_orig_score, run_b_rep_score, pbar=print_feedback),
                        'advanced': RMSE(self.run_a_orig_score, run_a_rep_score, pbar=print_feedback)}
            else:
                if print_feedback:
                    print("Determining Root Mean Square Error (RMSE) for baseline run.")
                with open(run_b_path, 'r') as b_run:
                    run_b_rep = pytrec_eval.parse_run(b_run)
                    run_b_rep = {t: run_b_rep[t] for t in sorted(run_b_rep)}
                    run_b_rep_score = self.rel_eval.evaluate(run_b_rep)
                return {'baseline': RMSE(self.run_b_orig_score, run_b_rep_score, pbar=print_feedback)}

        if self.run_b_orig_score and run_b_score:
            if self.run_a_orig_score and run_a_score:
                if print_feedback:
                    print("Determining Root Mean Square Error (RMSE) for baseline and advanced run.")
                return {'baseline': RMSE(self.run_b_orig_score, run_b_score, pbar=print_feedback),
                        'advanced': RMSE(self.run_a_orig_score, run_a_score, pbar=print_feedback)}
            else:
                if print_feedback:
                    print("Determining Root Mean Square Error (RMSE) for baseline run.")
                return {'baseline': RMSE(self.run_b_orig_score, run_b_score, pbar=print_feedback)}
        if self.run_b_orig_score and self.run_b_rep_score:
            if self.run_a_orig_score and self.run_a_rep_score:
                if print_feedback:
                    print("Determining Root Mean Square Error (RMSE) for baseline and advanced run.")
                return {'baseline': RMSE(self.run_b_orig_score, self.run_b_rep_score, pbar=print_feedback),
                        'advanced': RMSE(self.run_a_orig_score, self.run_a_rep_score, pbar=print_feedback)}
            else:
                if print_feedback:
                    print("Determining Root Mean Square Error (RMSE) for baseline run.")
                return {'baseline': RMSE(self.run_b_orig_score, self.run_b_rep_score, pbar=print_feedback)}
        else:
            print(ERR_MSG)

    def nrmse(self, run_b_score=None, run_a_score=None, run_b_path=None, run_a_path=None, print_feedback=False):
        """
        Determines the normalized Root Mean Square Error (RMSE).

        @param run_b_score: Scores of the baseline run,
                            if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_a_score: Scores of the advanced run,
                            if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_b_path: Path to another reproduced baseline run,
                           if not provided the reproduced baseline run of the RpdEvaluator object will be used instead.
        @param run_a_path: Path to another reproduced advanced run,
                           if not provided the reproduced advanced run of the RpdEvaluator object will be used instead.
        @param print_feedback: Boolean value indicating if feedback on progress should be printed.
        @return: Dictionary with nRMSE values that measure the closeness
                 between the topics scores of the original and reproduced runs.
        """
        if self.run_b_orig and run_b_path:
            if self.run_a_orig and run_a_path:
                if print_feedback:
                    print("Determining normalized Root Mean Square Error (RMSE) for baseline and advanced run.")
                with open(run_b_path, 'r') as b_run, open(run_a_path, 'r') as a_run:
                    run_b_rep = pytrec_eval.parse_run(b_run)
                    run_b_rep = {t: run_b_rep[t] for t in sorted(run_b_rep)}
                    run_b_rep_score = self.rel_eval.evaluate(run_b_rep)
                    run_a_rep = pytrec_eval.parse_run(a_run)
                    run_a_rep = {t: run_a_rep[t] for t in sorted(run_a_rep)}
                    run_a_rep_score = self.rel_eval.evaluate(run_a_rep)
                return {'baseline': nRMSE(self.run_b_orig_score, run_b_rep_score, pbar=print_feedback),
                        'advanced': nRMSE(self.run_a_orig_score, run_a_rep_score, pbar=print_feedback)}
            else:
                if print_feedback:
                    print("Determining normalized Root Mean Square Error (RMSE) for baseline run.")
                with open(run_b_path, 'r') as b_run:
                    run_b_rep = pytrec_eval.parse_run(b_run)
                    run_b_rep = {t: run_b_rep[t] for t in sorted(run_b_rep)}
                    run_b_rep_score = self.rel_eval.evaluate(run_b_rep)
                return {'baseline': nRMSE(self.run_b_orig_score, run_b_rep_score, pbar=print_feedback)}

        if self.run_b_orig_score and run_b_score:
            if self.run_a_orig_score and run_a_score:
                if print_feedback:
                    print("Determining normalized Root Mean Square Error (RMSE) for baseline and advanced run.")
                return {'baseline': nRMSE(self.run_b_orig_score, run_b_score, pbar=print_feedback),
                        'advanced': nRMSE(self.run_a_orig_score, run_a_score, pbar=print_feedback)}
            else:
                if print_feedback:
                    print("Determining normalized Root Mean Square Error (RMSE) for baseline run.")
                return {'baseline': nRMSE(self.run_b_orig_score, run_b_score, pbar=print_feedback)}
        if self.run_b_orig_score and self.run_b_rep_score:
            if self.run_a_orig_score and self.run_a_rep_score:
                if print_feedback:
                    print("Determining Root Mean Square Error (RMSE) for baseline and advanced run.")
                return {'baseline': nRMSE(self.run_b_orig_score, self.run_b_rep_score, pbar=print_feedback),
                        'advanced': nRMSE(self.run_a_orig_score, self.run_a_rep_score, pbar=print_feedback)}
            else:
                if print_feedback:
                    print("Determining normalized Root Mean Square Error (RMSE) for baseline run.")
                return {'baseline': nRMSE(self.run_b_orig_score, self.run_b_rep_score, pbar=print_feedback)}
        else:
            print(ERR_MSG)

    def ttest(self, run_b_score=None, run_a_score=None, run_b_path=None, run_a_path=None, print_feedback=False):
        """
        Conducts a paired two-tailed t-test for reproduced runs that were derived from the same test collection
        as in the original experiment.

        @param run_b_score: Scores of the baseline run,
                            if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_a_score: Scores of the advanced run,
                            if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_b_path: Path to another reproduced baseline run,
                           if not provided the reproduced baseline run of the RpdEvaluator object will be used instead.
        @param run_a_path: Path to another reproduced advanced run,
                           if not provided the reproduced advanced run of the RpdEvaluator object will be used instead.
        @param print_feedback: Boolean value indicating if feedback on progress should be printed.
        @return: Dictionary with p-values that compare the score distributions of the baseline and advanced run.
        """
        if run_b_path:
            if run_a_path:
                with open(run_b_path, 'r') as b_run, open(run_a_path, 'r') as a_run:
                    run_b_rep = pytrec_eval.parse_run(b_run)
                    run_b_rep = {t: run_b_rep[t] for t in sorted(run_b_rep)}
                    run_b_rep_score = self.rel_eval.evaluate(run_b_rep)
                    run_a_rep = pytrec_eval.parse_run(a_run)
                    run_a_rep = {t: run_a_rep[t] for t in sorted(run_a_rep)}
                    run_a_rep_score = self.rel_eval.evaluate(run_a_rep)
                return self._ttest(run_b_score=run_b_rep_score, run_a_score=run_a_rep_score, print_feedback=print_feedback)
            else:
                with open(run_b_path, 'r') as b_run:
                    run_b_rep = pytrec_eval.parse_run(b_run)
                    run_b_rep = {t: run_b_rep[t] for t in sorted(run_b_rep)}
                    run_b_rep_score = self.rel_eval.evaluate(run_b_rep)
                return self._ttest(run_b_score=run_b_rep_score, run_a_score=None, print_feedback=print_feedback)

        return self._ttest(run_b_score=run_b_score, run_a_score=run_a_score, print_feedback=print_feedback)


class RplEvaluator(Evaluator):
    """
    The Replicability Evaluator is used for quantifying the different levels of replication for runs that were
    derived from a test collection not used in the original experiment.
    """
    def __init__(self, **kwargs):
        super(RplEvaluator, self).__init__(**kwargs)
        self.qrel_rpl_path = kwargs.get('qrel_rpl_path', None)

        if self.qrel_rpl_path:
            with open(self.qrel_rpl_path, 'r') as f_qrel:
                qrel_rpl = pytrec_eval.parse_qrel(f_qrel)
                self.rel_eval_rpl = pytrec_eval.RelevanceEvaluator(qrel_rpl, pytrec_eval.supported_measures)

    def evaluate(self, run=None):
        """
        Evaluates the scores of the original and replicated baseline and advanced runs.
        If a (replicated) run is provided only this one will be evaluated and a dictionary with the corresponding
        scores is returned.
        @param run: A replicated run. If not specified, the original and replicated runs of the the RplEvaluator will
                    be used instead.
        @return: If run is specified, a dictionary with the corresponding scores is returned.
        """
        if run:
            return self.rel_eval_rpl.evaluate(run)

        super(RplEvaluator, self).evaluate()
        if self.run_b_rep:
            self.run_b_rep = break_ties(self.run_b_rep)
            self.run_b_rep_score = self.rel_eval_rpl.evaluate(self.run_b_rep)
        if self.run_a_rep:
            self.run_a_rep = break_ties(self.run_a_rep)
            self.run_a_rep_score = self.rel_eval_rpl.evaluate(self.run_a_rep)

    def ttest(self, run_b_score=None, run_a_score=None, run_b_path=None, run_a_path=None, print_feedback=False):
        """
        Conducts an un-paired two-tailed t-test for replicated runs that were derived from a test collection
        not used in the original experiment.

        @param run_b_score: Scores of the baseline run,
                            if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_a_score: Scores of the advanced run,
                            if not provided the scores of the RpdEvaluator object will be used instead.
        @param run_b_path: Path to another replicated baseline run,
                           if not provided the replicated baseline run of the RplEvaluator object will be used instead.
        @param run_a_path: Path to another replicated advanced run,
                           if not provided the replicated advanced run of the RplEvaluator object will be used instead.
        @param print_feedback: Boolean value indicating if feedback on progress should be printed.
        @return: Dictionary with p-values that compare the score distributions of the baseline and advanced run.
        """
        if run_b_path:
            if run_a_path:
                with open(run_b_path, 'r') as b_run, open(run_a_path, 'r') as a_run:
                    run_b_rep = pytrec_eval.parse_run(b_run)
                    run_b_rep = {t: run_b_rep[t] for t in sorted(run_b_rep)}
                    run_b_rep_score = self.rel_eval_rpl.evaluate(run_b_rep)
                    run_a_rep = pytrec_eval.parse_run(a_run)
                    run_a_rep = {t: run_a_rep[t] for t in sorted(run_a_rep)}
                    run_a_rep_score = self.rel_eval_rpl.evaluate(run_a_rep)
                return self._ttest(rpd=False, run_b_score=run_b_rep_score, run_a_score=run_a_rep_score, print_feedback=print_feedback)
            else:
                with open(run_b_path, 'r') as b_run:
                    run_b_rep = pytrec_eval.parse_run(b_run)
                    run_b_rep = {t: run_b_rep[t] for t in sorted(run_b_rep)}
                    run_b_rep_score = self.rel_eval_rpl.evaluate(run_b_rep)
                return self._ttest(rpd=False, run_b_score=run_b_rep_score, run_a_score=None, print_feedback=print_feedback)

        return self._ttest(rpd=False, run_b_score=run_b_score, run_a_score=run_a_score, print_feedback=print_feedback)



TRIM_THRESH = 1000  # default threshold for trimming the runs
PHI = 0.8  # default parameter for the Rank-Biased Overlap (RBO)
ERR_MSG = 'Please provide adequate run combinations and have them evaluated first.'  # error message

# evaluation measures of trec_eval that will be excluded from the reproduction and replication measures
exclude = [
    'runid',
    'num_q',
    'num_ret',
    'num_rel',
    'num_rel_ret',
    'num_nonrel_judged_ret',
    'relstring'
]



"""Evaluation measures at the level of effectiveness."""

import numpy as np
from math import sqrt
from copy import deepcopy
from tqdm import tqdm
from repro_eval.config import exclude


def _rmse(orig_score, rep_core, pbar=False):
    """
    Helping function returning a generator to determine the Root Mean Square Error (RMSE) for all topics.

    @param orig_score: The original scores.
    @param rep_core: The reproduced/replicated scores.
    @param pbar: Boolean value indicating if progress bar should be printed.
    @return: Generator with RMSE values.
    """
    orig_cp = deepcopy(orig_score)
    rep_cp = deepcopy(rep_core)
    measures_all = list(list(orig_cp.values())[0].keys())
    topics = orig_cp.keys()
    measures_valid = [m for m in measures_all if m not in exclude]

    measures = tqdm(measures_valid) if pbar else measures_valid

    for measure in measures:
        orig_measure = np.array([orig_cp.get(topic).get(measure) for topic in topics])
        rpl_measure = np.array([rep_cp.get(topic).get(measure) for topic in topics])
        diff = orig_measure - rpl_measure
        yield measure, sqrt(sum(np.square(diff))/len(diff))


def rmse(orig_score, rep_score, pbar=False):
    """
    Determines the Root Mean Square Error (RMSE) between the original and reproduced topic scores
    according to the following paper:
    Timo Breuer, Nicola Ferro, Norbert Fuhr, Maria Maistro, Tetsuya Sakai, Philipp Schaer, Ian Soboroff.
    How to Measure the Reproducibility of System-oriented IR Experiments.
    Proceedings of SIGIR, pages 349-358, 2020.

    @param orig_score: The original scores.
    @param rep_core: The reproduced/replicated scores.
    @param pbar: Boolean value indicating if progress bar should be printed.
    @return: Dictionary with RMSE values that measure the closeness between the original and reproduced topic scores.
    """
    return dict(_rmse(orig_score, rep_score, pbar=pbar))


def _maxrmse(orig_score, pbar=False):
    """
    Helping function returning a generator to determine the maximum Root Mean Square Error (RMSE) for all topics.

    @param orig_score: The original scores.
    @param pbar: Boolean value indicating if progress bar should be printed.
    @return: Generator with RMSE values.
    """
    orig_cp = deepcopy(orig_score)
    measures_all = list(list(orig_cp.values())[0].keys())
    topics = orig_cp.keys()
    measures_valid = [m for m in measures_all if m not in exclude]
    measures = tqdm(measures_valid) if pbar else measures_valid

    for measure in measures:
        orig_measure = np.array([orig_cp.get(topic).get(measure) for topic in topics])
        _max = np.vectorize(lambda x: max(x, 1 - x))
        maxdiff = _max(orig_measure)
        yield measure, sqrt(sum(np.square(maxdiff))/len(maxdiff))


def nrmse(orig_score, rep_score, pbar=False):
    """
    Determines the normalized Root Mean Square Error (RMSE) between the original and reproduced topic scores.

    @param orig_score: The original scores.
    @param rep_core: The reproduced/replicated scores.
    @param pbar: Boolean value indicating if progress bar should be printed.
    @return: Dictionary with RMSE values that measure the closeness between the original and reproduced topic scores.
    """
    rmse = dict(_rmse(orig_score, rep_score, pbar=pbar))
    maxrmse = dict(_maxrmse(orig_score, pbar=pbar))
    return {measure: score / maxrmse.get(measure) for measure, score in rmse.items()}



"""Evaluation measures at the level of document orderings."""

from repro_eval.config import TRIM_THRESH, PHI
from scipy.stats.stats import kendalltau
from tqdm import tqdm
from repro_eval.measure.external.rbo import rbo
from repro_eval.util import break_ties


def _rbo(run, ideal, p, depth):
    # Implementation taken from the TREC Health Misinformation Track with modifications
    # see also: https://github.com/claclark/Compatibility
    run_set = set()
    ideal_set = set()

    score = 0.0
    normalizer = 0.0
    weight = 1.0
    for i in range(depth):
        if i < len(run):
            run_set.add(run[i])
        if i < len(ideal):
            ideal_set.add(ideal[i])
        score += weight*len(ideal_set.intersection(run_set))/(i + 1)
        normalizer += weight
        weight *= p
    return score/normalizer


def _ktau_union(orig_run, rep_run, trim_thresh=TRIM_THRESH, pbar=False):
    """
    Helping function returning a generator to determine Kendall's tau Union (KTU) for all topics.

    @param orig_run: The original run.
    @param rep_run: The reproduced/replicated run.
    @param trim_thresh: Threshold values for the number of documents to be compared.
    @param pbar: Boolean value indicating if progress bar should be printed.
    @return: Generator with KTU values.
    """

    generator = tqdm(rep_run.items()) if pbar else rep_run.items()

    for topic, docs in generator:
        orig_docs = list(orig_run.get(topic).keys())[:trim_thresh]
        rep_docs = list(rep_run.get(topic).keys())[:trim_thresh]
        union = list(sorted(set(orig_docs + rep_docs)))
        orig_idx = [union.index(doc) for doc in orig_docs]
        rep_idx = [union.index(doc) for doc in rep_docs]
        yield topic, round(kendalltau(orig_idx, rep_idx).correlation, 14)


def ktau_union(orig_run, rep_run, trim_thresh=TRIM_THRESH, pbar=False):
    """
    Determines the Kendall's tau Union (KTU) between the original and reproduced document orderings
    according to the following paper:
    Timo Breuer, Nicola Ferro, Norbert Fuhr, Maria Maistro, Tetsuya Sakai, Philipp Schaer, Ian Soboroff.
    How to Measure the Reproducibility of System-oriented IR Experiments.
    Proceedings of SIGIR, pages 349-358, 2020.

    @param orig_run: The original run.
    @param rep_run: The reproduced/replicated run.
    @param trim_thresh: Threshold values for the number of documents to be compared.
    @param pbar: Boolean value indicating if progress bar should be printed.
    @return: Dictionary with KTU values that compare the document orderings of the original and reproduced runs.
    """

    # Safety check for runs that are not added via pytrec_eval
    orig_run = break_ties(orig_run)
    rep_run = break_ties(rep_run)

    return dict(_ktau_union(orig_run, rep_run, trim_thresh=trim_thresh, pbar=pbar))


def _RBO(orig_run, rep_run, phi, trim_thresh=TRIM_THRESH, pbar=False, misinfo=True):
    """
    Helping function returning a generator to determine the Rank-Biased Overlap (RBO) for all topics.

    @param orig_run: The original run.
    @param rep_run: The reproduced/replicated run.
    @param phi: Parameter for top-heaviness of the RBO.
    @param trim_thresh: Threshold values for the number of documents to be compared.
    @param pbar: Boolean value indicating if progress bar should be printed.
    @param misinfo: Use the RBO implementation that is also used in the TREC Health Misinformation Track.
                    See also: https://github.com/claclark/Compatibility
    @return: Generator with RBO values.
    """

    generator = tqdm(rep_run.items()) if pbar else rep_run.items()

    if misinfo:
        for topic, docs in generator:
            yield topic, _rbo(list(rep_run.get(topic).keys())[:trim_thresh],
                              list(orig_run.get(topic).keys())[:trim_thresh],
                              p=phi,
                              depth=trim_thresh)

    else:
        for topic, docs in generator:
            yield topic, rbo(list(rep_run.get(topic).keys())[:trim_thresh],
                             list(orig_run.get(topic).keys())[:trim_thresh],
                             p=phi).ext


def RBO(orig_run, rep_run, phi=PHI, trim_thresh=TRIM_THRESH, pbar=False, misinfo=True):
    """
    Determines the Rank-Biased Overlap (RBO) between the original and reproduced document orderings
    according to the following paper:
    Timo Breuer, Nicola Ferro, Norbert Fuhr, Maria Maistro, Tetsuya Sakai, Philipp Schaer, Ian Soboroff.
    How to Measure the Reproducibility of System-oriented IR Experiments.
    Proceedings of SIGIR, pages 349-358, 2020.

    @param orig_run: The original run.
    @param rep_run: The reproduced/replicated run.
    @param phi: Parameter for top-heaviness of the RBO.
    @param trim_thresh: Threshold values for the number of documents to be compared.
    @param pbar: Boolean value indicating if progress bar should be printed.
    @param misinfo: Use the RBO implementation that is also used in the TREC Health Misinformation Track.
                    See also: https://github.com/claclark/Compatibility
    @return: Dictionary with RBO values that compare the document orderings of the original and reproduced runs.
    """

    # Safety check for runs that are not added via pytrec_eval
    orig_run = break_ties(orig_run)
    rep_run = break_ties(rep_run)

    return dict(_RBO(orig_run, rep_run, phi=phi, trim_thresh=trim_thresh, pbar=pbar, misinfo=misinfo))



import os
import json
import platform
import pkg_resources
import warnings
from collections import defaultdict
from io import BytesIO, TextIOWrapper
import cpuinfo
import pytrec_eval
import git
from ruamel.yaml import YAML
from repro_eval import Evaluator

META_START = '# ir_metadata.start'
META_END = '# ir_metadata.end'

class PrimadExperiment:
    """
    The PrimadExperiment is used to determine the reproducibility measures 
    between a reference run and a set of one or more reproduced run files.
    Depending on the type of the PRIMAD experiment, several reproducibility 
    measures can be determined.
    
    @param ref_base_path: Path to a single run file that corresponds to the 
                          original (or reference) baseline of the experiments. 
    @param ref_adv_path: Path to a single run file that corresponds to the
                         original (or reference) baseline of the experiments.
    @param primad: String with lower and upper case letters depending on which 
                   PRIMAD components have changed in the experiments, e.g., 
                   "priMad" when only the Method changes due to parameter sweeps.
    @param rep_base: List containing paths to run files that reproduce the 
                     original (or reference) baseline run.
    @param rpd_qrels: Qrels file that is used to evaluate the reproducibility of
                      the experiments, i.e., it used to evaluate runs that are 
                      derived from the same test collection.
    @param rep_adv: List containing paths to run files that reproduce the 
                    original (or reference) advanced run.
    @param rpl_qrels: Qrels file that is used to evaluate the replicability of
                      the experiments, i.e., it is used to evaluate runs that are 
                      derived from a different test collection. Please note that 
                      "rpd_qrels" has to be provided too.
    """
    def __init__(self, **kwargs): 
    
        self.ref_base_path = kwargs.get('ref_base_path', None)
        if self.ref_base_path:
            self.ref_base_run = MetadataHandler.strip_metadata(self.ref_base_path)
        else:
            self.ref_base_run = None
        
        self.ref_adv_path = kwargs.get('ref_adv_path', None)
        if self.ref_adv_path:
            self.ref_adv_run = MetadataHandler.strip_metadata(self.ref_adv_path)
        else:
            self.ref_adv_run = None    
        
        self.primad = kwargs.get('primad', None)
        self.rep_base = kwargs.get('rep_base', None)
        self.rpd_qrels = kwargs.get('rpd_qrels', None)
        self.rep_adv = kwargs.get('rep_adv', None)
        self.rpl_qrels = kwargs.get('rpl_qrels', None)

        if self.rpl_qrels:
            self.rep_eval = Evaluator.RplEvaluator(qrel_orig_path=self.rpd_qrels,
                                                   qrel_rpl_path=self.rpl_qrels)  
            
            with open(self.rpd_qrels, 'r') as f_rpd_qrels, open(self.rpl_qrels, 'r') as f_rpl_qrels:
                qrels = pytrec_eval.parse_qrel(f_rpd_qrels)
                self.rpd_rel_eval = pytrec_eval.RelevanceEvaluator(qrels, pytrec_eval.supported_measures)
                qrels = pytrec_eval.parse_qrel(f_rpl_qrels)
                self.rpl_rel_eval = pytrec_eval.RelevanceEvaluator(qrels, pytrec_eval.supported_measures)
            
        
        elif self.primad[-1].islower(): # check if data component is the same
            self.rep_eval = Evaluator.RpdEvaluator(qrel_orig_path=self.rpd_qrels)
            
            with open(self.rpd_qrels, 'r') as f_qrels:
                qrels = pytrec_eval.parse_qrel(f_qrels)
                self.rpd_rel_eval = pytrec_eval.RelevanceEvaluator(qrels, pytrec_eval.supported_measures)
            
        else:
            raise ValueError('Please provide a correct combination of qrels and PRIMAD type.')
            
    def get_primad_type(self):
        """
        This method returns a string that identifies the type of the 
        PRIMAD experiment.

        @return: String with lower and upper case letters depending on which 
                 PRIMAD components have changed in the experiments, e.g., 
                 "priMad" when only the Method changes due to parameter sweeps.
        """
        return self.primad
    
    def evaluate(self):
        """
        This method validates the PRIMAD experiment in accordance with the given
        "primad" identifier. Currently, the following experiments are supported.
            - priMad: Parameter sweep
            - PRIMAd: Reproducibility evaluation on the same test collection
            - PRIMAD: Generalizability evaluation
        
        @return: Dictionary containing the average retrieval performance and 
                 the reproducibility measures for each run.
        """
        
        if self.primad == 'priMad':
            if self.ref_adv_run is None and self.rep_adv is None:
                
                evaluations = {} 
                
                self.rep_eval.run_b_orig = self.ref_base_run
                self.rep_eval.evaluate()

                for rep_run_path in self.rep_base + [self.ref_base_path]:
                    
                    run_evaluations = {}
                
                    rep_run = MetadataHandler.strip_metadata(rep_run_path)
                    scores = self.rpd_rel_eval.evaluate(rep_run)
                    
                    run_evaluations['arp'] = scores
                    run_evaluations['ktu'] = self.rep_eval.ktau_union(run_b_rep=rep_run).get('baseline')
                    run_evaluations['rbo'] = self.rep_eval.rbo(run_b_rep=rep_run).get('baseline')
                    run_evaluations['rmse'] = self.rep_eval.nrmse(run_b_score=scores).get('baseline')
                    run_evaluations['pval'] = self.rep_eval.ttest(run_b_score=scores).get('baseline')
                
                    run_name = os.path.basename(rep_run_path)
                
                    evaluations[run_name] = run_evaluations
        
                return evaluations
    
        if self.primad == 'PRIMAd':
            
            evaluations = {} 
            
            self.rep_eval.run_b_orig = self.ref_base_run
            self.rep_eval.run_a_orig = self.ref_adv_run
            self.rep_eval.trim(t=1000)
            self.rep_eval.evaluate()
    
            pairs = self._find_pairs(rep_base=self.rep_base, rep_adv=self.rep_adv)
            pairs = pairs + [{'base': self.ref_base_path, 'adv': self.ref_adv_path}]
            
            for pair in pairs:
                
                pair_evaluations = {}
                
                rep_run_base = MetadataHandler.strip_metadata(pair.get('base'))
                rep_meta_base = MetadataHandler.read_metadata(pair.get('base'))                
                rep_run_adv = MetadataHandler.strip_metadata(pair.get('adv'))
                rep_meta_adv = MetadataHandler.read_metadata(pair.get('adv'))          
                      
                self.rep_eval.trim(t=1000, run=rep_run_base)
                self.rep_eval.trim(t=1000, run=rep_run_adv)
                scores_base = self.rpd_rel_eval.evaluate(rep_run_base)
                scores_adv = self.rpd_rel_eval.evaluate(rep_run_adv)
                arp = {'baseline': scores_base, 'advanced': scores_adv}
                pair_evaluations['arp'] = arp 
                pair_evaluations['ktu'] = self.rep_eval.ktau_union(run_b_rep=rep_run_base, run_a_rep=rep_run_adv)
                pair_evaluations['rbo'] = self.rep_eval.rbo(run_b_rep=rep_run_base, run_a_rep=rep_run_adv)
                pair_evaluations['rmse'] = self.rep_eval.nrmse(run_b_score=scores_base, run_a_score=scores_adv)
                pair_evaluations['er'] = self.rep_eval.er(run_b_score=scores_base, run_a_score=scores_adv)
                pair_evaluations['dri'] = self.rep_eval.dri(run_b_score=scores_base, run_a_score=scores_adv)
                pair_evaluations['pval'] = self.rep_eval.ttest(run_b_score=scores_base, run_a_score=scores_adv)
                
                if rep_meta_base.get('actor').get('team') == rep_meta_adv.get('actor').get('team'):
                    expid = rep_meta_base.get('actor').get('team')
                else:
                    expid = '_'.join([rep_meta_base.get('tag'), rep_meta_adv.get('tag')])
                
                evaluations[expid] = pair_evaluations
 
            return evaluations
    
        if self.primad == 'PRIMAD':
            evaluations = {} 
            
            self.rep_eval.run_b_orig = self.ref_base_run
            self.rep_eval.run_a_orig = self.ref_adv_run
            self.rep_eval.trim(t=1000)
            self.rep_eval.evaluate()
    
            pairs = self._find_pairs(rep_base=self.rep_base, rep_adv=self.rep_adv)
            pairs = pairs
            
            for pair in pairs:
                
                pair_evaluations = {}
                
                rep_run_base = MetadataHandler.strip_metadata(pair.get('base'))
                rep_meta_base = MetadataHandler.read_metadata(pair.get('base'))                
                rep_run_adv = MetadataHandler.strip_metadata(pair.get('adv'))
                rep_meta_adv = MetadataHandler.read_metadata(pair.get('adv'))          
                      
                self.rep_eval.trim(t=1000, run=rep_run_base)
                self.rep_eval.trim(t=1000, run=rep_run_adv)
                scores_base = self.rpl_rel_eval.evaluate(rep_run_base)
                scores_adv = self.rpl_rel_eval.evaluate(rep_run_adv)
                arp = {'baseline': scores_base, 'advanced': scores_adv}
                pair_evaluations['arp'] = arp 
                pair_evaluations['er'] = self.rep_eval.er(run_b_score=scores_base, run_a_score=scores_adv)
                pair_evaluations['dri'] = self.rep_eval.dri(run_b_score=scores_base, run_a_score=scores_adv)
                pair_evaluations['pval'] = self.rep_eval.ttest(run_b_score=scores_base, run_a_score=scores_adv)
                
                expid = '_'.join([rep_meta_base.get('tag'), rep_meta_adv.get('tag')])  
                evaluations[expid] = pair_evaluations
            
            return evaluations
        
        else:
            raise ValueError('The specified type of the PRIMAD experiments is not supported yet.')
        
    def _find_pairs(self, rep_base, rep_adv):
        """
        This method finds pairs between lists of baseline and advanced runs. 
        A pair is defined by the highest number of matching PRIMAD components.
        
        @param rep_base: List with baseline runs.
        @param rep_adv: List with advanced runs.
        
        @return: List with dictionaries containing paths to a baseline and an 
                 advanced run.
        """

        pairs = []
        for brp in rep_base:
            br = MetadataHandler.read_metadata(run_path=brp)

            arp = None
            cnt = 0
            
            for _arp in rep_adv:
                _cnt = 0
                ar = MetadataHandler.read_metadata(run_path=_arp)
                
                for k,v in br.items():
                    if v == ar.get(k):
                        _cnt += 1
                        
                if _cnt > cnt:
                    cnt = _cnt 
                    arp = _arp
            
            pairs.append({'base': brp, 'adv': arp})
            
        return pairs
    

class MetadataAnalyzer:
    """
    The MetadataAnalyzer is used to analyze set of different run files in
    reference to a run that has be be provided upon instantiation. The 
    analyze_directory() method returns a dictionary with PRIMAD identifiers as
    keys and lists with the corresponding run paths as values.
    
    @param run_path: Path to the reference run file.
    """
    
    def __init__(self, run_path):
        
        self.reference_run_path = run_path
        self.reference_run = MetadataHandler.strip_metadata(run_path)
        self.reference_metadata = MetadataHandler.read_metadata(run_path)
        
    def set_reference(self, run_path):
        """
        Use this method to set a new reference run.
        
        @param run_path: Path to the new reference run file.
        """
        
        self.reference_run_path = run_path
        self.reference_run = MetadataHandler.strip_metadata(run_path)
        self.reference_metadata = MetadataHandler.read_metadata(run_path)
        
    def analyze_directory(self, dir_path):    
        """
        Use this method to analyze the specified directory in comparison to the 
        reference run. 
        
        @param dir_path: Path to the directory.
        """
        
        components = ['platform', 'research goal', 'implementation', 'method', 'actor', 'data']
        primad = {}
        
        files = os.listdir(dir_path)
        
        for _file in files:
            file_path = os.path.join(dir_path, _file)
    
            if file_path == self.reference_run_path:
                continue
            
            _metadata = MetadataHandler.read_metadata(file_path)
            
            primad_str = ''
                                        
            for component in components:
                if self.reference_metadata[component] != _metadata[component]:
                    primad_str += component[0].upper()
                else:
                    primad_str += component[0]
            
            primad[file_path] = primad_str
            
        experiments = defaultdict(list)
        for k, v in primad.items(): 
            experiments[v].append(k)
            
        return experiments  
    
    @staticmethod
    def filter_by_baseline(ref_run, runs):
        """
        Use this method to filter a list of runs wrt. to the baseline that is 
        specified under "research goal/evaluation/baseline" of a given reference run.
        
        @param ref_run: The reference with the baseline.
        @param runs: A list of run paths that is filtered.
        """
        
        run_tag = MetadataHandler.read_metadata(ref_run).get('tag')

        filtered_list = []
        for run in runs:
            _metadata = MetadataHandler.read_metadata(run)
            baseline = _metadata.get('research goal').get('evaluation').get('baseline')[0]
            if baseline == run_tag:
                filtered_list.append(run)
                
        return filtered_list
    
    @staticmethod 
    def filter_by_test_collection(test_collection, runs):
        """
        Use this method to filter a list of runs wrt. to the test collection
        specified under "data/test_collection".
        
        @param test_collection: Name of the test collection.
        @param runs: A list of run paths that is filtered.
        """
        
        filtered_list = []
        for run in runs:
            _metadata = MetadataHandler.read_metadata(run)
            name = _metadata.get('data').get('test collection').get('name')
            if test_collection == name:
                filtered_list.append(run)
                
        return filtered_list


class MetadataHandler:
    """
    Use the MetadataHandler for in- and output operations of annotated run files.
    
    @param run_path: Path the run file without metadata annotations. It is also 
                     possible to load an already annotated run and modify it with
                     the MetadataHandler.
    @param metadata_path: Path to the YAML file containing the metadata that 
                          should be added to the run file.
    """
    def __init__(self, run_path, metadata_path=None):
         
        self.run_path = run_path
        
        if metadata_path:
            self._metadata = MetadataHandler.read_metadata_template(metadata_path)
        else:
            self._metadata = MetadataHandler.read_metadata(run_path)
        
    def get_metadata(self):
        """
        Use this method to get the currently set metadata annotations.
        
        @return: Nested dictionary containing the metadata annotations.
        """
        
        return self._metadata
    
    def set_metadata(self, metadata_dict=None, metadata_path=None):
        """
        Use this method to set/update the metadata. It can either be provided with a 
        dictionary of a path to a YAML file
        
        @param metadata_dict: Nested dictionary containing the metadata annotations.
        @param metadata_path: Path to the YAML file with metadata.
        """
        if metadata_path:
            self._metadata = MetadataHandler.read_metadata_template(metadata_path)
        
        if metadata_dict:
            self._metadata = metadata_dict
    
    def dump_metadata(self, dump_path=None, complete_metadata=False, repo_path='.'):
        """
        Use this method to dump the current metadata into a YAML file. 
        The filename is a concatenation of the run tag and the "_annotated" suffix.
        
        @param dump_path: Path to the directory where the metadata is dumped.
        @param complete_metadata: If true, the Platform and Implementation will 
                                  be added automatically, if not already provided.
        @param repo_path: Path to the git repository of the Implementation that 
                          underlies the run file. This path is needed for the 
                          automatic completion.
        """
           
        if complete_metadata:
            self.complete_metadata(repo_path=repo_path)
            
        if self._metadata:
                
            tag = self._metadata['tag'] 
            f_out_name = '_'.join([tag, 'dump.yaml'])  
            f_out_path = os.path.join(dump_path, f_out_name)    
                
            with open(f_out_path, 'wb') as f_out:
                bytes_io = BytesIO()
                yaml = YAML()
                yaml.width = 4096
                yaml.dump(self._metadata, bytes_io)
                f_out.write(bytes_io.getvalue())

    def write_metadata(self, run_path=None, complete_metadata=False, repo_path='.'):
        """
        This method writes the metadata into the run file.
        
        @param run_path: Path to the annotated run file.
        @param complete_metadata: If true, the Platform and Implementation will 
                                  be added automatically, if not already provided.
        @param repo_path: Path to the git repository of the Implementation that 
                          underlies the run file. This path is needed for the 
                          automatic completion.
        """
        if complete_metadata:
            self.complete_metadata(repo_path=repo_path)
        
        bytes_io = BytesIO()
        yaml = YAML()
        yaml.width = 4096
        yaml.dump(self._metadata, bytes_io)
    
        byte_str = bytes_io.getvalue().decode('UTF-8')
        lines = byte_str.split('\n')
        
        if run_path is None:
            f_out_path = '_'.join([self.run_path, 'annotated'])
        else:
            f_out_path = '_'.join([run_path])
        
        with open(f_out_path, 'w') as f_out:
            
            f_out.write(''.join([META_START, '\n']))
            for line in lines[:-1]:
                f_out.write(' '.join(['#', line, '\n']))            
            f_out.write(''.join([META_END, '\n']))
            
            with open(self.run_path, 'r') as f_in:
                for run_line in f_in.readlines():
                    f_out.write(run_line)
                
    def complete_metadata(self, repo_path='.'):
        """
        This method automatically adds metadata about the Platform and 
        the Implementation component.
        
        @param repo_path: Path to the git repository of the Implementation that 
                          underlies the run file. If not specified this method
                          assumes that the program is executed from the root 
                          directory of the git repository.
        """
        if self._metadata.get('platform') is None:
            platform_dict = {
                'hardware': {
                    'cpu': self._get_cpu(),  
                    'ram': self._get_ram(),
                    },
                'operating system': self._get_os(),
                'software': self._get_libs(),
                }
            
            self._metadata['platform'] = platform_dict
            
        if self._metadata.get('implementation') is None: 
            self._metadata['implementation'] = self._get_src(repo_path=repo_path)
    
    @staticmethod
    def strip_metadata(annotated_run):
        '''
        Strips off the metadata and returns a dict-version of the run that is parsed with pytrec_eval.
        
        @param annotated_run: Path to the annotated run file.
        
        @return: defaultdict that can be used with pytrec_eval or repro_eval.
        '''

        with TextIOWrapper(buffer=BytesIO(), encoding='utf-8', line_buffering=True) as text_io_wrapper:
            with open(annotated_run, 'r') as f_in:
                lines = f_in.readlines()
                for line in lines:
                    if line[0] != '#':
                        text_io_wrapper.write(line)
            text_io_wrapper.seek(0,0)        
            run = pytrec_eval.parse_run(text_io_wrapper)
                        
        return run

    @staticmethod
    def read_metadata(run_path):
        '''
        Reads the metadata out of an annotated run and returns a dict containing the metadata.
        
        @param run_path: Path to the run file.
        
        @return: Dictionary containing the metadata information of the annotated 
                 run file.
        '''
        
        _metadata = None
        
        with open(run_path, 'r') as f_in: 
            lines = f_in.readlines()
            if lines[0].strip('\n') == META_START:
                metadata_str = ''
                yaml=YAML(typ='safe')

                for line in lines[1:]:
                    if line.strip('\n') != META_END:
                        metadata_str += line.strip('#')
                    else:
                        break
                _metadata = yaml.load(metadata_str)
        
        return _metadata
    
    @staticmethod
    def read_metadata_template(metadata_path):
        """
        This method reads in a YAML file containing the metadata.
        
        @param template_path: Path to the metadata YAML file.
        
        @return: Nested dictionary containing the metadata.
        """
        
        with open(metadata_path, 'r') as f_in:
            yaml = YAML(typ='safe')
            return yaml.load(f_in)

    def _get_cpu(self):
        """
        Reads out metadata information about the CPU including the model's name
        the architectures, the operation mode and the number of available cores.
        """
        
        cpu = cpuinfo.get_cpu_info()
        return {
            'model': cpu['brand_raw'],
            'architecture': platform.machine(),
            'operation mode': '-'.join([str(cpu['bits']), 'bit']),
            'number of cores': cpu['count'],
        }
      
    def _get_os(self):
        """
        Reads out metadata information about the operating system including 
        the platform (e.g. Linux), the kernel release version, 
        and the distribution's name.
        """
        
        try:
            with open("/etc/os-release") as f_in:
                os_info = {}
                for line in f_in:
                    k,v = line.rstrip().split('=')
                    os_info[k] = v.strip('"')
                
            distribution = os_info['PRETTY_NAME']
            
        except:
            warnings.warn('/etc/os-release not found. Using the available information of the platform package instead.')
            distribution = platform.version()
             
        return {
            'platform': platform.system(),
            'kernel': platform.release(),
            'distribution': distribution,
        }

    def _get_ram(self):
        """
        Reads out the available RAM and returns the size in GB.
        """
        
        memory_bytes = os.sysconf('SC_PAGE_SIZE') * os.sysconf('SC_PHYS_PAGES') 
        memory_gb = memory_bytes/(1024.0 ** 3)  
        return ' '.join([str(round(memory_gb, 2)),'GB'])

    def _get_libs(self):
        """
        Reads out all installed Python packages of the active environment.
        """
        
        installed_packages = [d.project_name for d in pkg_resources.working_set]
        return {'libraries': {'python': installed_packages}}

    def _get_src(self, repo_path='.'):
        """ 
        Reads out information from the specified repository.
        
        @param repo_path: Path to the git repository of the Implementation that 
                          underlies the run file. If not specified this method
                          assumes that the program is executed from the root 
                          directory of the git repository.        
        """
        
        extensions_path = pkg_resources.resource_filename(__name__, 'resources/extensions.json')

        repo = git.Repo(repo_path)
        
        with open(extensions_path, 'r') as input_file:
            extensions = json.load(input_file)
            
        languages = set()

        for _, _, files in os.walk('.'):
            for name in files:
                _, file_extension = os.path.splitext(name)
                language = extensions.get(file_extension[1:])
                if language:
                    languages.add(language)

        return {
            'repository': repo.remote().url,
            'commit': str(repo.head.commit),
            'lang': list(languages),
        }



import math
from scipy.stats.stats import ttest_rel, ttest_ind
from tqdm import tqdm
from repro_eval.util import topic_scores


def _ttest(orig_score, rep_score, rpd=True, pbar=False):
    """

    @param orig_score: The original scores.
    @param rep_score: The reproduced/replicated scores.
    @param rpd: Boolean indicating if the evaluated runs are reproduced.
    @param pbar: Boolean value indicating if progress bar should be printed.
    @return: Generator with p-values.
    """
    if rpd:  # paired two-tailed t-test
        topic_scores_orig = topic_scores(orig_score)
        topic_scores_rep = topic_scores(rep_score)

        generator = tqdm(topic_scores_orig.items()) if pbar else topic_scores_orig.items()

        for measure, scores in generator:
            yield measure, ttest_rel(scores, topic_scores_rep.get(measure)).pvalue

    else:  # else unpaired two-tailed t-test
        topic_scores_orig = topic_scores(orig_score)
        topic_scores_rep = topic_scores(rep_score)

        generator = tqdm(topic_scores_orig.items()) if pbar else topic_scores_orig.items()

        for measure, scores in generator:
            yield measure, ttest_ind(scores, topic_scores_rep.get(measure)).pvalue


def ttest(orig_score, rep_score, rpd=True, pbar=False):
    """

    @param orig_score: The original scores.
    @param rep_score: The reproduced/replicated scores.
    @param rpd: Boolean indicating if the evaluated runs are reproduced.
    @param pbar: Boolean value indicating if progress bar should be printed.
    @return: Dictionary with p-values that compare the score distributions of the baseline and advanced run.
    """
    pvals = dict(_ttest(orig_score, rep_score, rpd=rpd, pbar=pbar))
    nan_list = list(filter(lambda x: math.isnan(x), pvals.values()))
    if len(nan_list) == len(pvals):  # is every pval is nan?
        if orig_score == rep_score:  # equal score distributions?
            pvals = dict.fromkeys(pvals, 1.0)

    return pvals



import numpy as np
from copy import deepcopy
from tqdm import tqdm
from repro_eval.config import exclude


def diff(topic_score_a, topic_score_b):
    """
    Use this function to get a generator with absoulte differences
    between the topic scores of the baseline and advanced runs.

    @param topic_score_a: Topic scores of the advanced run.
    @param topic_score_b: Topic scores of the baseline run.
    @return: Generator with absolute differences between the topics scores.
    """
    for measure, value in topic_score_a.items():
        if measure not in exclude:
            yield measure, value - topic_score_b.get(measure)


def topic_diff(run_a, run_b):
    """
    Use this function to get a generator with absoulte differences
    between the topic scores of the baseline and advanced runs for each measure.

    @param run_a: The advanced run.
    @param run_b: The baseline run.
    @return: Generator with absolute differences between the topics scores for each measure.
    """
    run_a_cp = deepcopy(run_a)
    run_b_cp = deepcopy(run_b)

    for topic, measures in run_a_cp.items():
        yield topic, dict(diff(measures, run_b_cp.get(topic)))


def _mean_improvement(run_a, run_b):
    """
    Helping function returning a generator for determining the mean improvements.

    @param run_a: The advanced run.
    @param run_b: The baseline run.
    @return: Generator with mean improvements.
    """
    measures_all = list(list(run_a.values())[0].keys())
    measures_valid = [m for m in measures_all if m not in exclude]
    topics = run_a.keys()
    delta = dict(topic_diff(run_a, run_b))

    for measure in measures_valid:
        yield measure, np.array([delta.get(topic).get(measure) for topic in topics]).mean()


def mean_improvement(run_a, run_b):
    """
    Determines the relative improvement that is used to derive the Delta Relative Improvement (DeltaRI).

    @param run_a: The advanced run.
    @param run_b: The baseline run.
    @return: Dictionary with mean improvements for each measure.
    """
    return dict(_mean_improvement(run_a, run_b))


def _er(orig_score_a, orig_score_b, rep_score_a, rep_score_b, pbar=False):
    """
    Helping function returning a generator for determining the Effect Ratio (ER).

    @param orig_score_a: Scores of the original advanced run.
    @param orig_score_b: Scores of the original baseline run.
    @param rep_score_a: Scores of the reproduced/replicated advanced run.
    @param rep_score_b: Scores of the reproduced/replicated baseline run.
    @param pbar: Boolean value indicating if progress bar should be printed.
    @return: Generator with ER scores.
    """
    mi_orig = mean_improvement(orig_score_a, orig_score_b)
    mi_rep = mean_improvement(rep_score_a, rep_score_b)

    generator = tqdm(mi_rep.items()) if pbar else mi_rep.items()

    for measure, value in generator:
        yield measure, value / mi_orig.get(measure)


def ER(orig_score_a, orig_score_b, rep_score_a, rep_score_b, pbar=False):
    """
    Determines the Effect Ratio (ER) according to the following paper:
    Timo Breuer, Nicola Ferro, Norbert Fuhr, Maria Maistro, Tetsuya Sakai, Philipp Schaer, Ian Soboroff.
    How to Measure the Reproducibility of System-oriented IR Experiments.
    Proceedings of SIGIR, pages 349-358, 2020.

    The ER value is determined by the ratio between the mean improvements
    of the original and reproduced/replicated experiments.

    @param orig_score_a: Scores of the original advanced run.
    @param orig_score_b: Scores of the original baseline run.
    @param rep_score_a: Scores of the reproduced/replicated advanced run.
    @param rep_score_b: Scores of the reproduced/replicated baseline run.
    @param pbar: Boolean value indicating if progress bar should be printed.
    @return: Dictionary containing the ER values for the specified run combination.
    """
    return dict(_er(orig_score_a, orig_score_b, rep_score_a, rep_score_b, pbar=pbar))


def _mean_score(scores):
    """
    Helping function to determine the mean scores across the topics for each measure.

    @param scores: Run scores.
    @return: Generator with mean scores.
    """
    measures_all = list(list(scores.values())[0].keys())
    measures_valid = [m for m in measures_all if m not in exclude]
    topics = scores.keys()

    for measure in measures_valid:
        yield measure, np.array([scores.get(topic).get(measure) for topic in topics]).mean()


def mean_score(scores):
    """
    Use this function to compute the mean scores across the topics for each measure.

    @param scores: Run scores.
    @return: Dictionary containing the mean scores for each measure.
    """
    return dict(_mean_score(scores))


def _rel_improve(scores_a, scores_b):
    """
    Helping function returning a generator for determining the relative improvements.

    @param scores_a: Scores of the advanced run.
    @param scores_b: Scores of the baseline run.
    @return: Generator with relative improvements.
    """
    mean_scores_a = mean_score(scores_a)
    mean_scores_b = mean_score(scores_b)

    for measure, mean in mean_scores_a.items():
        yield measure, (mean - mean_scores_b.get(measure)) / mean_scores_b.get(measure)


def rel_improve(scores_a, scores_b):
    """
    Determines the relative improvement that is used to derive the Delta Relative Improvement (DeltaRI).

    @param scores_a: Scores of the advanced run.
    @param scores_b: Scores of the baseline run.
    @return: Dictionary with relative improvements for each measure.
    """
    return dict(_rel_improve(scores_a, scores_b))


def _deltaRI(orig_score_a, orig_score_b, rep_score_a, rep_score_b, pbar=False):
    """
    Helping function returning a generator for determining the Delta Relative Improvement (DeltaRI).

    @param orig_score_a: Scores of the original advanced run.
    @param orig_score_b: Scores of the original baseline run.
    @param rep_score_a: Scores of the reproduced/replicated advanced run.
    @param rep_score_b: Scores of the reproduced/replicated baseline run.
    @param pbar: Boolean value indicating if progress bar should be printed.
    @return: Generator with DeltaRI scores.
    """
    rel_improve_orig = rel_improve(orig_score_a, orig_score_b)
    rel_improve_rep = rel_improve(rep_score_a, rep_score_b)

    generator = tqdm(rel_improve_orig.items()) if pbar else rel_improve_orig.items()

    for measure, ri in generator:
        yield measure, ri - rel_improve_rep.get(measure)


def deltaRI(orig_score_a, orig_score_b, rep_score_a, rep_score_b, pbar=False):
    """
    Determines the Delta Relative Improvement (DeltaRI) according to the following paper:
    Timo Breuer, Nicola Ferro, Norbert Fuhr, Maria Maistro, Tetsuya Sakai, Philipp Schaer, Ian Soboroff.
    How to Measure the Reproducibility of System-oriented IR Experiments.
    Proceedings of SIGIR, pages 349-358, 2020.

    The DeltaRI value is determined by the difference between the relative improvements
    of the original and reproduced/replicated experiments.

    @param orig_score_a: Scores of the original advanced run.
    @param orig_score_b: Scores of the original baseline run.
    @param rep_score_a: Scores of the reproduced/replicated advanced run.
    @param rep_score_b: Scores of the reproduced/replicated baseline run.
    @param pbar: Boolean value indicating if progress bar should be printed.
    @return: Dictionary containing the DeltaRI values for the specified run combination.
    """
    return dict(_deltaRI(orig_score_a, orig_score_b, rep_score_a, rep_score_b, pbar=pbar))



import pathlib
from setuptools import setup

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(name='repro_eval',
      version='0.4.0',
      description='A tool to quantify the replicability and reproducibility of system-oriented IR experiments.',
      long_description=README,
      long_description_content_type="text/markdown",
      url='http://github.com/irgroup/repro_eval',
      author='Timo Breuer',
      author_email='timo.breuer@th-koeln.de',
      license='MIT',
      packages=['repro_eval', 
                'repro_eval.measure', 
                'repro_eval.measure.external'],
      install_requires=[
          'pytrec_eval',
          'numpy',
          'scipy',
          'tqdm',
          'ruamel.yaml',
          'GitPython',
          'py-cpuinfo'
      ],
      include_package_data=True,
      package_data={'': ['resources/*.json']},
      zip_safe=False)



"""
Use repro_eval from the command line with e.g.

python -m repro_eval -t rpd -q qrel_orig -r orig_b rpd_b

python -m repro_eval -t rpd -q qrel_orig -r orig_b orig_a rpd_b rpd_a

python -m repro_eval -t rpd -m rmse -q qrel_orig -r orig_b rpd_b

python -m repro_eval -t rpl -q qrel_orig qrel_rpl -r orig_b rpl_b

python -m repro_eval -t rpl -q qrel_orig qrel_rpl -r orig_b orig_a rpl_b rpl_a

after having installed the Python package.
For other more specific examples also have a look at the README file.
Depending on the provided parameters and input run files,
evaluation measures will be printed.
"""

import argparse
from repro_eval.Evaluator import RpdEvaluator, RplEvaluator
from repro_eval.util import print_simple_line, print_base_adv
from repro_eval.util import arp


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument('-t', '--type')
    parser.add_argument('-m', '--measure', nargs='+')
    parser.add_argument('-q', '--qrels', nargs='+')
    parser.add_argument('-r', '--runs', nargs='+')

    args = parser.parse_args()

    if args.type in ['rpd', 'reproducibility']:
        if len(args.runs) == 4:
            rpd_eval = RpdEvaluator(qrel_orig_path=args.qrels[0],
                                    run_b_orig_path=args.runs[0],
                                    run_a_orig_path=args.runs[1],
                                    run_b_rep_path=args.runs[2],
                                    run_a_rep_path=args.runs[3])

        if len(args.runs) == 2:
            rpd_eval = RpdEvaluator(qrel_orig_path=args.qrels[0],
                                    run_b_orig_path=args.runs[0],
                                    run_a_orig_path=None,
                                    run_b_rep_path=args.runs[1],
                                    run_a_rep_path=None)

        rpd_eval.trim()
        rpd_eval.evaluate()

        measure_list = args.measure if args.measure is not None else []

        # KTU
        if 'ktu' in measure_list or args.measure is None:
            ktu = rpd_eval.ktau_union()
            print("Kendall's tau Union (KTU)")
            print('------------------------------------------------------------------')
            for topic, value in ktu.get('baseline').items():
                value_adv = ktu.get('advanced').get(topic) if ktu.get('advanced') is not None else None
                print_base_adv(topic, 'KTU', value, value_adv)
            value_adv = arp(ktu.get('advanced')) if ktu.get('advanced') is not None else None
            print_base_adv('ARP', 'KTU', arp(ktu.get('baseline')), value_adv)
            print()

        # RBO
        if 'rbo' in measure_list or args.measure is None:
            rbo = rpd_eval.rbo()
            print("Rank-biased Overlap (RBO)")
            print('------------------------------------------------------------------')
            for topic, value in rbo.get('baseline').items():
                value_adv = rbo.get('advanced').get(topic) if rbo.get('advanced') is not None else None
                print_base_adv(topic, 'RBO', value, value_adv)
            value_adv = arp(rbo.get('advanced')) if rbo.get('advanced') is not None else None
            print_base_adv('ARP', 'RBO', arp(rbo.get('baseline')), value_adv)
            print()

        # RMSE
        if 'rmse' in measure_list or args.measure is None:
            rmse = rpd_eval.rmse()
            print("Root mean square error (RMSE)")
            print('------------------------------------------------------------------')
            for measure, value in rmse.get('baseline').items():
                value_adv = rmse.get('advanced').get(measure) if rmse.get('advanced') is not None else None
                print_base_adv(measure, 'RMSE', value, value_adv)
            print()

        # ER
        if 'er' in measure_list or args.measure is None and len(args.runs) == 4:
            print("Effect ratio (ER)")
            print('------------------------------------------------------------------')
            er = rpd_eval.er()
            for measure, value in er.items():
                print_simple_line(measure, 'ER', value)
            print()

        # DRI
        if 'dri' in measure_list or args.measure is None and len(args.runs) == 4:
            print("Delta Relative Improvement (DRI)")
            print('------------------------------------------------------------------')
            dri = rpd_eval.dri()
            for measure, value in dri.items():
                print_simple_line(measure, 'DRI', value)
            print()

        # ttest
        if 'ttest' in measure_list or args.measure is None:
            pvals = rpd_eval.ttest()
            print("Two-tailed paired t-test (p-value)")
            print('------------------------------------------------------------------')
            for measure, value in pvals.get('baseline').items():
                value_adv = pvals.get('advanced').get(measure) if pvals.get('advanced') is not None else None
                print_base_adv(measure, 'PVAL', value, value_adv)
            print()

    if args.type in ['rpl', 'replicability']:
        if len(args.runs) == 4:
            rpl_eval = RplEvaluator(qrel_orig_path=args.qrels[0],
                                    run_b_orig_path=args.runs[0],
                                    run_a_orig_path=args.runs[1],
                                    run_b_rep_path=args.runs[2],
                                    run_a_rep_path=args.runs[3],
                                    qrel_rpl_path=args.qrels[1])

        if len(args.runs) == 2:
            rpl_eval = RplEvaluator(qrel_orig_path=args.qrels[0],
                                    run_b_orig_path=args.runs[0],
                                    run_a_orig_path=None,
                                    run_b_rep_path=args.runs[1],
                                    run_a_rep_path=None,
                                    qrel_rpl_path=args.qrels[1])

        rpl_eval.trim()
        rpl_eval.evaluate()

        measure_list = args.measure if args.measure is not None else []

        # ER
        if 'er' in measure_list or args.measure is None and len(args.runs) == 4:
            print("Effect ratio (ER)")
            print('------------------------------------------------------------------')
            er = rpl_eval.er()
            for measure, value in er.items():
                print_simple_line(measure, 'ER', value)
            print()

        # DRI
        if 'dri' in measure_list or args.measure is None and len(args.runs) == 4:
            print("Delta Relative Improvement (DRI)")
            print('------------------------------------------------------------------')
            dri = rpl_eval.dri()
            for measure, value in dri.items():
                print_simple_line(measure, 'DRI', value)
            print()

        # ttest
        if 'ttest' in measure_list or args.measure is None:
            pvals = rpl_eval.ttest()
            print("Two-tailed unpaired t-test (p-value)")
            print('------------------------------------------------------------------')
            for measure, value in pvals.get('baseline').items():
                value_adv = pvals.get('advanced').get(measure) if pvals.get('advanced') is not None else None
                print_base_adv(measure, 'PVAL', value, value_adv)
            print()


if __name__ == "__main__":
    main()
